from typing import Optional

import tensorflow as tf

from tensorflow_projection_qm.metrics.metric import LocalizableMetric
from tensorflow_projection_qm.util import distance


@tf.function
def neighborhood_hit_impl(X_2d, y, k):
    D_low = distance.psqdist(X_2d)
    _, topk_ixs = distance.nearest_k(D_low, k)

    return tf.reduce_mean(tf.cast(tf.gather(y, topk_ixs) == y[:, tf.newaxis], tf.float64), -1)


def neighborhood_hit(X_2d, y, k):
    return tf.reduce_mean(neighborhood_hit_impl(X_2d, y, tf.constant(k)))


def neighborhood_hit_with_local(X_2d, y, k) -> tuple[tf.Tensor, tf.Tensor]:
    per_point = neighborhood_hit_impl(X_2d, y, tf.constant(k))
    return tf.reduce_mean(per_point), per_point


class NeighborhoodHit(LocalizableMetric):
    name = "neighborhood_hit"

    def __init__(self, k: Optional[int] = None) -> None:
        super().__init__()
        self.k = k

    @property
    def config(self):
        return {"k": self.k}

    def measure(self, X_2d, y):
        if self._with_local:
            return neighborhood_hit_with_local(X_2d, y, self.k)
        return neighborhood_hit(X_2d, y, self.k)

    def measure_from_dict(self, args: dict):
        return self.measure(args["X_2d"], args["y"])



import tensorflow as tf

from tensorflow_projection_qm.metrics.metric import Metric
from tensorflow_projection_qm.util import distance


@tf.function
def pearson_correlation_impl(X, X_2d):
    flat_d_high = tf.sqrt(distance.flat_psqdist(X))
    flat_d_low = tf.sqrt(distance.flat_psqdist(X_2d))

    mean_high = tf.reduce_mean(flat_d_high, axis=-1)
    mean_low = tf.reduce_mean(flat_d_low, axis=-1)

    diff_mean_high = flat_d_high - mean_high
    diff_mean_low = flat_d_low - mean_low

    return tf.reduce_sum(
        tf.math.l2_normalize(diff_mean_high, axis=-1) * tf.math.l2_normalize(diff_mean_low, axis=-1)
    )


def pearson_correlation(X, X_2d):
    return pearson_correlation_impl(X, X_2d)


class PearsonCorrelation(Metric):
    name = "pearson_correlation"

    def __init__(self) -> None:
        super().__init__()

    @property
    def config(self):
        return {}

    def measure(self, X, X_2d):
        return pearson_correlation(X, X_2d)

    def measure_from_dict(self, args: dict):
        return self.measure(args["X"], args["X_2d"])



from tensorflow_projection_qm.metrics.continuity import class_aware_continuity

import numpy as np
import numpy.testing as npt


def test_single_data_point():
    X = np.array([[1.0, 2.0, 3.0]])  # a single data point in 3-D
    X_2d = np.array([[0.0, -1.0]])  # a single projected data point
    y = np.array([0])
    npt.assert_equal(class_aware_continuity(X, X_2d, y, k=1, n_classes=1).numpy(), 1.0)


def test_single_data_point_first_class_isnt_zero():
    X = np.array([[1.0, 2.0, 3.0]])  # a single data point in 3-D
    X_2d = np.array([[0.0, -1.0]])  # a single projected data point
    y = np.array([5])
    npt.assert_equal(class_aware_continuity(X, X_2d, y, k=1, n_classes=1).numpy(), 1.0)


def test_two_data_points_same_class():
    X = np.array([[1.0, 1.0, 1.0], [0.0, 0.0, 1.0]])
    X_2d = np.array([[0.0, 0.0], [10.0, 10.0]])
    y = np.array([0, 0])
    # In this case, one data point is always the other's nearest
    # neighbor, irrespective of the distance. Continuity should
    # be perfect (== 1.0)
    npt.assert_equal(class_aware_continuity(X, X_2d, y, k=1, n_classes=1).numpy(), 1.0)

    # This should not care whether their classes differ or not.
    y = np.array([0, 1])
    npt.assert_equal(class_aware_continuity(X, X_2d, y, k=1, n_classes=2).numpy(), 1.0)

    y = np.array([3, 3])
    npt.assert_equal(class_aware_continuity(X, X_2d, y, k=1, n_classes=2).numpy(), 1.0)

    y = np.array([3, 5])
    npt.assert_equal(class_aware_continuity(X, X_2d, y, k=1, n_classes=2).numpy(), 1.0)


def test_one_error_diff_class():
    # data in a line: 0 -- 1 - 2
    X = np.c_[np.zeros(3), np.array([0.0, 1.5, 2.0]), np.zeros(3)]
    # data in a line, with one swap: 0 -- 2 - 1
    X_2d = np.c_[np.zeros(3), np.array([0.0, 2.0, 1.2])]
    y = np.array([1, 0, 0])
    # Nearest neighbor:
    #    Data Space | Proj Space |
    # 0 |     1     |     2      |  Penalty = 0 because across diff classes
    # 1 |     2     |     2      |  Penalty = 0
    # 2 |     1     |     1      |  Penalty = 0
    npt.assert_allclose(class_aware_continuity(X, X_2d, y, k=1, n_classes=2).numpy(), 1.0)


def test_one_error_same_class():
    # data in a line: 0 -- 1 - 2
    X = np.c_[np.zeros(3), np.array([0.0, 1.5, 2.0]), np.zeros(3)]
    # data in a line, with one swap: 0 -- 2 - 1
    X_2d = np.c_[np.zeros(3), np.array([0.0, 2.0, 1.2])]
    y = np.array([0, 0, 3])
    # Nearest neighbor:
    #    Data Space | Proj Space |
    # 0 |     1     |     2      |  Penalty = 1
    # 1 |     2     |     2      |  Penalty = 0
    # 2 |     1     |     1      |  Penalty = 0
    npt.assert_allclose(class_aware_continuity(X, X_2d, y, k=1, n_classes=2).numpy(), 1 - 1.0 / 3.0)


def test_two_errors():
    # data in a line: 0 -- 1 - 2
    X = np.c_[np.zeros(3), np.array([0.0, 1.5, 2.0]), np.zeros(3)]
    # data in a line, with one swap: 0 - 2 -- 1
    X_2d = np.c_[np.zeros(3), np.array([0.0, 2.0, 0.9])]
    # Nearest neighbor:
    #    Data Space | Proj Space |
    # 0 |     1     |     2      |  Penalty = 1
    # 1 |     2     |     2      |  Penalty = 0
    # 2 |     1     |     0      |  Penalty = 1

    # No penalty applies because all classes are different
    y = np.array([0, 1, 2])
    npt.assert_allclose(class_aware_continuity(X, X_2d, y, k=1, n_classes=3).numpy(), 1.0)

    # One penalty applies
    y = np.array([0, 0, 2])
    npt.assert_allclose(class_aware_continuity(X, X_2d, y, k=1, n_classes=3).numpy(), 1 - 1.0 / 3.0)

    # Two penalties apply
    y = np.array([0, 0, 0])
    npt.assert_allclose(class_aware_continuity(X, X_2d, y, k=1, n_classes=3).numpy(), 1 - 2.0 / 3.0)


def test_k_larger_than_dataset_size():
    X = np.random.randn(10, 5)
    X_2d = np.random.randn(10, 2)
    y = np.zeros(10)
    # If k > n-1 (here n == 10), then we're out of neighbors to use for the
    # computation. The calculation should adapt accordingly (i.e., k shouldn't
    # have an effect on the result).
    results = [
        class_aware_continuity(X, X_2d, y, k=_k, n_classes=1).numpy() for _k in (10, 20, 1000)
    ]

    npt.assert_allclose(results[0], results[1])
    npt.assert_allclose(results[1], results[2])



from tensorflow_projection_qm.metrics.continuity import continuity

import numpy as np
import numpy.testing as npt


def test_single_data_point():
    X = np.array([[1.0, 2.0, 3.0]])  # a single data point in 3-D
    X_2d = np.array([[0.0, -1.0]])  # a single projected data point

    npt.assert_equal(continuity(X, X_2d, k=1).numpy(), 1.0)


def test_two_data_points():
    X = np.array([[1.0, 1.0, 1.0], [0.0, 0.0, 1.0]])
    X_2d = np.array([[0.0, 0.0], [10.0, 10.0]])

    # In this case, one data point is always the other's nearest
    # neighbor, irrespective of the distance. Continuity should
    # be perfect (== 1.0)
    npt.assert_equal(continuity(X, X_2d, k=1).numpy(), 1.0)


def test_one_error():
    # data in a line: 0 -- 1 - 2
    X = np.c_[np.zeros(3), np.array([0.0, 1.5, 2.0]), np.zeros(3)]
    # data in a line, with one swap: 0 -- 2 - 1
    X_2d = np.c_[np.zeros(3), np.array([0.0, 2.0, 1.2])]

    # Nearest neighbor:
    #    Data Space | Proj Space |
    # 0 |     1     |     2      |  Penalty = 1
    # 1 |     2     |     2      |  Penalty = 0
    # 2 |     1     |     1      |  Penalty = 0
    npt.assert_allclose(continuity(X, X_2d, k=1).numpy(), 1 - 1.0 / 3.0)


def test_two_errors():
    # data in a line: 0 -- 1 - 2
    X = np.c_[np.zeros(3), np.array([0.0, 1.5, 2.0]), np.zeros(3)]
    # data in a line, with one swap: 0 - 2 -- 1
    X_2d = np.c_[np.zeros(3), np.array([0.0, 2.0, 0.9])]

    # Nearest neighbor:
    #    Data Space | Proj Space |
    # 0 |     1     |     2      |  Penalty = 1
    # 1 |     2     |     2      |  Penalty = 0
    # 2 |     1     |     0      |  Penalty = 1
    npt.assert_allclose(continuity(X, X_2d, k=1).numpy(), 1 - (1 + 1) / 3.0)


def test_k_larger_than_dataset_size():
    X = np.random.randn(10, 5)
    X_2d = np.random.randn(10, 2)

    # If k > n-1 (here n == 10), then we're out of neighbors to use for the
    # computation. The calculation should adapt accordingly (i.e., k shouldn't
    # have an effect on the result).
    results = [continuity(X, X_2d, k=_k).numpy() for _k in (10, 20, 1000)]

    npt.assert_allclose(results[0], results[1])
    npt.assert_allclose(results[1], results[2])


def test_no_errors_possible():
    # data in a line: 0 -- 1 - 2
    X = np.c_[np.zeros(3), np.array([0.0, 1.5, 2.0]), np.zeros(3)]
    # data in a line, with one swap: 0 - 2 -- 1
    X_2d = np.c_[np.zeros(3), np.array([0.0, 2.0, 0.9])]

    # Since K = n-1, there are no possible errors, continuity is perfect.

    # Nearest neighbors:
    #    Data Space | Proj Space |
    # 0 |    1, 2   |     2, 1   |  Penalty = 0
    # 1 |    2, 0   |     2, 0   |  Penalty = 0
    # 2 |    1, 0   |     0, 1   |  Penalty = 0
    npt.assert_allclose(continuity(X, X_2d, k=2).numpy(), 1)


def test_two_missing_neighbors():
    # data in a line: 0 -- 1 - 2 -- 3
    X = np.c_[np.zeros(4), np.array([0.0, 1.5, 2.0, 3.0]), np.zeros(4)]
    # data in a line, with one swap: 0 - 2 -- 1 - 3
    X_2d = np.c_[np.zeros(4), np.array([0.0, 2.0, 0.9, 2.1])]

    # Nearest neighbors:
    #    Data Space | Proj Space |
    # 0 |    1, 2   |     2, 1   |  Penalty = 0
    # 1 |    2, 0   |     3, 2   |  Penalty = 3 - 2 = 1 (0 is missing neighbor)
    # 2 |    1, 3   |     0, 1   |  Penalty = 3 - 2 = 1 (3 is missing neighbor)
    # 3 |    2, 1   |     1, 2   |  Penalty = 0
    npt.assert_allclose(continuity(X, X_2d, k=2).numpy(), 1 - 2.0 / 4.0)


def test_three_missing_neighbors_diff_ranks():
    # data: 0 -- 1 - 2 -- 3
    #            |
    #            4
    # proj: 0 --- 1 - 2 -- 3 ---- 4
    X = np.c_[np.zeros(5), np.array([0.0, 1.5, 2.0, 3.1, 1.5]), np.array([0, 0, 0, 0, 1.0])]
    X_2d = np.c_[np.zeros(5), np.array([0.0, 1.5, 2.0, 3.1, 10.0])]

    # Nearest neighbors:
    #    Data Space | Proj Space | Diff |
    # 0 |    1, 4   |     1, 2   |  [4] | Penalty = 4 - 2 = 2 (4 is missing)
    # 1 |    2, 4   |     2, 0   |  [4] | Penalty = 4 - 2 = 2 (4 is missing)
    # 2 |    1, 3   |     1, 3   |  []  | Penalty = 0
    # 3 |    2, 1   |     2, 1   |  []  | Penalty = 0
    # 4 |    1, 2   |     3, 2   |  [1] | Penalty = 3 - 2 = 1 (1 is missing)
    npt.assert_allclose(
        continuity(X, X_2d, k=2).numpy(),
        np.mean(1 - np.array([2.0, 2.0, 0.0, 0.0, 1.0]) / (2 * 5 - 3 * 2 - 1)),
    )



import tensorflow as tf

from tensorflow_projection_qm.metrics.metric import Metric
from tensorflow_projection_qm.util import distance


@tf.function
def shepard_goodness_impl(X, X_2d):
    flat_d_high = distance.flat_psqdist(X)
    flat_d_low = distance.flat_psqdist(X_2d)
    n = tf.cast(tf.shape(flat_d_high)[0], tf.float32)

    ranks_high = tf.argsort(tf.argsort(flat_d_high))
    ranks_low = tf.argsort(tf.argsort(flat_d_low))

    # Assuming no ties, we can use this formula.
    return 1 - 6 * tf.cast(
        tf.reduce_sum(tf.math.squared_difference(ranks_high, ranks_low)), tf.float32
    ) / (n * (n**2 - 1))


def shepard_goodness(X, X_2d):
    return shepard_goodness_impl(X, X_2d)


class ShepardGoodness(Metric):
    name = "shepard_goodness"

    def __init__(self) -> None:
        super().__init__()

    @property
    def config(self):
        return {}

    def measure(self, X, X_2d):
        return shepard_goodness(X, X_2d)

    def measure_from_dict(self, args: dict):
        return self.measure(args["X"], args["X_2d"])



import tensorflow as tf

from tensorflow_projection_qm.metrics.metric import LocalizableMetric
from tensorflow_projection_qm.util import distance


@tf.function
def average_local_error_impl(X, X_2d):
    D_high = distance.psqdist(X)
    D_low = distance.psqdist(X_2d)

    D_high /= tf.reduce_max(D_high)
    D_low /= tf.reduce_max(D_low)

    n = tf.shape(D_high)[0]
    avg_local_err = (1 / (n - 1)) * tf.reduce_sum(tf.abs(D_high - D_low), -1)

    return avg_local_err


def average_local_error(X, X_2d):
    return tf.reduce_mean(average_local_error_impl(X, X_2d))


def average_local_error_with_local(X, X_2d):
    per_point = average_local_error_impl(X, X_2d)
    return tf.reduce_mean(per_point), per_point


class AverageLocalError(LocalizableMetric):
    name = "average_local_error"

    def __init__(self) -> None:
        super().__init__()

    @property
    def config(self):
        return {}

    def measure(self, X, X_2d):
        if self._with_local:
            return average_local_error_with_local(X, X_2d)
        return average_local_error(X, X_2d)

    def measure_from_dict(self, args: dict):
        return self.measure(args["X"], args["X_2d"])



"""
Module implementing Mean Relative Rank Error metrics

The metrics are implemented as described in Nonlinear Dimensionality Reduction,
by Lee and Verleysen (2007), Chapter 6.
"""

from typing import Optional

import tensorflow as tf

from tensorflow_projection_qm.metrics.metric import LocalizableMetric
from tensorflow_projection_qm.util import distance


@tf.function
def mrre_data_impl(X, X_2d, k):
    n = tf.shape(X)[0]
    D_high = distance.psqdist(X)
    D_low = distance.psqdist(X_2d)

    nn_orig = distance.sort_distances(D_high)
    ixs_orig = tf.argsort(nn_orig)
    nn_proj = distance.sort_distances(D_low)
    knn_proj = nn_proj[:, 1 : k + 1]

    orig_ranks = tf.gather(ixs_orig, knn_proj, batch_dims=-1)
    unnormalized = tf.reduce_sum(
        tf.abs(orig_ranks - tf.expand_dims(tf.range(1, k + 1), 0))
        / tf.expand_dims(tf.range(1, k + 1), 0),
        axis=-1,
    )

    C = tf.reduce_sum(tf.abs(2 * tf.range(1, k + 1) - n - 1) / tf.range(1, k + 1))

    return (1 / C) * unnormalized


@tf.function
def mrre_proj_impl(X, X_2d, k):
    n = tf.shape(X)[0]
    D_high = distance.psqdist(X)
    D_low = distance.psqdist(X_2d)

    nn_orig = distance.sort_distances(D_high)
    knn_orig = nn_orig[:, 1 : k + 1]
    nn_proj = distance.sort_distances(D_low)
    ixs_proj = tf.argsort(nn_proj)

    proj_ranks = tf.gather(ixs_proj, knn_orig, batch_dims=-1)
    unnormalized = tf.reduce_sum(
        tf.abs(proj_ranks - tf.expand_dims(tf.range(1, k + 1), 0))
        / tf.expand_dims(tf.range(1, k + 1), 0),
        axis=-1,
    )

    C = tf.reduce_sum(tf.abs(2 * tf.range(1, k + 1) - n - 1) / tf.range(1, k + 1))

    return (1 / C) * unnormalized


def mrre_data_with_local(X, X_2d, k):
    per_point = mrre_data_impl(X, X_2d, tf.constant(k))
    return tf.reduce_mean(per_point), per_point


def mrre_proj_with_local(X, X_2d, k):
    per_point = mrre_proj_impl(X, X_2d, tf.constant(k))
    return tf.reduce_mean(per_point), per_point


def mrre_data(X, X_2d, k):
    """Compute the MRRE considering the data-space neighborhoods as ground truth.

    This corresponds to Equation 6.4 in Nonlinear Dimensionality Reduction.

    Args:
        X (tf.Tensor, np.ndarray): the data matrix for high-dimensional data
        X_2d (tf.Tensor, np.ndarray): the data matrix for projected data
        k (int): the neighborhood size to consider.

    Returns:
        tf.Tensor: a Tensor containing a single scalar, the metric value.
    """
    return tf.reduce_mean(mrre_data_impl(X, X_2d, tf.constant(k)))


def mrre_proj(X, X_2d, k):
    """Compute the MRRE considering the projection-space neighborhoods as ground truth.

    This corresponds to Equation 6.3 in Nonlinear Dimensionality Reduction.

    Args:
        X (tf.Tensor, np.ndarray): the data matrix for high-dimensional data
        X_2d (tf.Tensor, np.ndarray): the data matrix for projected data
        k (int): the neighborhood size to consider.

    Returns:
        tf.Tensor: a Tensor containing a single scalar, the metric value.
    """
    return tf.reduce_mean(mrre_proj_impl(X, X_2d, tf.constant(k)))


class MRREData(LocalizableMetric):
    name = "mrre_data"

    def __init__(self, k: Optional[int] = None) -> None:
        super().__init__()
        self.k = k

    @property
    def config(self):
        return {"k": self.k}

    def measure(self, X, X_2d):
        if self._with_local:
            return mrre_data_with_local(X, X_2d, self.k)
        return mrre_data(X, X_2d, self.k)

    def measure_from_dict(self, args: dict):
        return self.measure(args["X"], args["X_2d"])


class MRREProj(LocalizableMetric):
    name = "mrre_proj"

    def __init__(self, k: Optional[int] = None) -> None:
        super().__init__()
        self.k = k

    @property
    def config(self):
        return {"k": self.k}

    def measure(self, X, X_2d):
        if self._with_local:
            return mrre_proj_with_local(X, X_2d, self.k)
        return mrre_proj(X, X_2d, self.k)

    def measure_from_dict(self, args: dict):
        return self.measure(args["X"], args["X_2d"])



import tensorflow as tf

from tensorflow_projection_qm.metrics.metric import Metric
from tensorflow_projection_qm.util import distance


@tf.function
def raw_stress_from_distances(D_high, D_low):
    return tf.reduce_sum((D_high - D_low) ** 2)


@tf.function
def raw_stress(X, X_2d):
    D_high = tf.sqrt(distance.psqdist(X))
    D_low = tf.sqrt(distance.psqdist(X_2d))

    return raw_stress_from_distances(D_high, D_low)


@tf.function
def normalized_stress_from_distances(D_high, D_low):
    return tf.reduce_sum((D_high - D_low) ** 2) / tf.reduce_sum(D_high**2)


@tf.function
def normalized_stress(X, X_2d):
    D_high = tf.sqrt(distance.psqdist(X))
    D_low = tf.sqrt(distance.psqdist(X_2d))

    return normalized_stress_from_distances(D_high, D_low)


@tf.function
def scaled_stress_from_distances(D_high, D_low, alpha=1.0):
    return normalized_stress_from_distances(D_high, alpha * D_low)


@tf.function
def scaled_stress(X, X_2d, alpha=1.0):
    return normalized_stress(X, alpha * X_2d)


@tf.function
def scale_normalized_stress_from_distances(D_high, D_low):
    opt_alpha = tf.reduce_sum(D_high * D_low) / tf.reduce_sum(D_low**2)

    return normalized_stress_from_distances(D_high, opt_alpha * D_low)


@tf.function
def scale_normalized_stress(X, X_2d):
    """A scale-invariant version of the Stress metric.

    From the paper "Normalized Stress" is Not Normalized: How to Interpret Stress Correctly,
    by Smelser et al.

    Args:
        X (tf.Tensor, np.ndarray): The original data matrix.
        X_2d (tf.Tensor, np.ndarray): The projected data matrix.

    Returns:
        tf.Tensor: a tensor containing a single scalar, the value of the metric.
    """
    D_high = tf.sqrt(distance.psqdist(X))
    D_low = tf.sqrt(distance.psqdist(X_2d))

    return scale_normalized_stress_from_distances(D_high, D_low)


class NormalizedStress(Metric):
    name = "normalized_stress"

    def __init__(self) -> None:
        super().__init__()

    @property
    def config(self):
        return {}

    def measure(self, X, X_2d):
        return normalized_stress(X, X_2d)

    def measure_from_dict(self, args: dict):
        return self.measure(args["X"], args["X_2d"])


class ScaleNormalizedStress(Metric):
    name = "scale_normalized_stress"

    def __init__(self) -> None:
        super().__init__()

    @property
    def config(self):
        return {}

    def measure(self, X, X_2d):
        return scale_normalized_stress(X, X_2d)

    def measure_from_dict(self, args: dict):
        return self.measure(args["X"], args["X_2d"])



import pytest
from tensorflow_projection_qm.util import distance

import numpy as np
import numpy.testing as npt
from sklearn.metrics import pairwise_distances


class TestPSQDist:
    def test_singleton(self):
        A = np.array([[1.0]])

        D = distance.psqdist(A).numpy()
        npt.assert_array_equal(D, 0.0)

    def test_1d_data(self):
        A = np.array([[3.0], [1.0], [2.0]])

        my_res = distance.psqdist(A).numpy()

        npt.assert_(my_res.shape == (3, 3))
        npt.assert_array_equal(
            my_res,
            [
                [0.0, 4.0, 1.0],
                [4.0, 0.0, 1.0],
                [1.0, 1.0, 0.0],
            ],
        )

    def test_2d_data(self):
        A = np.array([[1.0, 2.0], [2.0, 1.0], [0.0, 0.0], [-1.0, -0.5]])

        my_res = distance.psqdist(A).numpy()

        npt.assert_(my_res.shape == (4, 4))
        npt.assert_array_equal(
            my_res,
            [
                [0.0, 2.0, 5.0, 2**2 + 2.5**2],
                [2.0, 0.0, 5.0, 3**2 + 1.5**2],
                [5.0, 5.0, 0.0, 1.25],
                [2**2 + 2.5**2, 3**2 + 1.5**2, 1.25, 0.0],
            ],
        )

    def test_big_matrix(self):
        A = np.arange(1000 * 250).reshape(1000, 250)
        my_res = distance.psqdist(A).numpy()
        sklearn_res = pairwise_distances(A, metric="sqeuclidean")

        npt.assert_(my_res.shape == sklearn_res.shape == (1000, 1000))
        npt.assert_array_equal(my_res, sklearn_res)

    def test_big_random_matrix(self):
        A = np.random.randn(1000, 1000)
        my_res = distance.psqdist(A).numpy()
        sklearn_res = pairwise_distances(A, metric="sqeuclidean")
        npt.assert_(my_res.shape == sklearn_res.shape == (1000, 1000))
        npt.assert_array_almost_equal(my_res, sklearn_res)

        # Assert *strictly* equal fails, but the KNN should be the same
        my_nns = np.argsort(my_res)
        sklearn_nns = np.argsort(sklearn_res)

        npt.assert_array_equal(my_nns, sklearn_nns)


class TestCSQDist:
    def test_handles_singleton(self):
        out = distance.csqdist(np.array([[1.0]]), np.array([[2.0], [3.0], [4.0]]))

        npt.assert_allclose(out, np.array([[1.0, 4.0, 9.0]]))

    _DATA_MATRICES = [
        np.array([[3.0, -2.0, 1.0]]),
        np.array([[1.0, 2.0, -1.0], [2.0, 0.0, -3.0]]),
        np.array([[1.0, 2.0, -1.0], [2.0, 0.0, -3.0], [-2.5, 1.3, 2.6]]),
    ]

    @pytest.mark.parametrize("X", _DATA_MATRICES)
    @pytest.mark.parametrize("Y", _DATA_MATRICES)
    def test_handles_different_number_of_rows(self, X, Y):
        try:
            npt.assert_(distance.csqdist(X, Y).numpy().shape, (X.shape[0], Y.shape[0]))
        except Exception:
            pytest.fail()

    @pytest.mark.parametrize(
        "X",
        [
            np.array([[1.0]]),
            np.array([[3.0, 1.0, 1.0]]),
            np.array([[1.0, 2.0], [3.0, -4.0]]),
            np.arange(10_000).reshape(100, 100),
        ],
    )
    def test_compatible_with_psqdist(self, X):
        npt.assert_allclose(distance.csqdist(X, X), distance.psqdist(X))

    @pytest.mark.parametrize(
        "X,Y",
        [
            (np.array([[1.0]]), np.array([[2.0]])),
            (np.array([[-2.0, 0.0]]), np.array([[0.0, 0.0]])),
            (np.array([[-2.0, 0.0], [1.0, -1.0]]), np.array([[1.0, 2.0]])),
        ],
    )
    def test_compatible_with_sklearn(self, X, Y):
        npt.assert_allclose(distance.csqdist(X, Y), pairwise_distances(X, Y, metric="sqeuclidean"))



import tensorflow as tf


@tf.function
def sort_distances(D) -> tf.Tensor:
    """Sorts a distance matrix using tf.argsort with a corrective factor

    This function adds a "correction" to D, ensuring that, O_i0 = i for all i
    (where O is the output of this function). This is not guaranteed by
    tf.argsort for matrices where D_ij is 0 in a given row i for more than a
    single j. However, since we use distance matrices for k-nearest neighbors
    computations, it is common to assume that the first nearest neighbor of
    element i is itself (hence the need for this function).
    This function internally creates an *invalid* distance matrix
    due to the negative entries added to it, but it prevents such invalid
    matrices from leaking out to the caller.

    >>> D = tf.convert_to_tensor([
    ...         [0.0, 1.0, 0.0],
    ...         [1.0, 0.0, 0.5],
    ...         [0.0, 0.5, 0.0]], dtype=tf.float64)
    >>> sort_distances(D).numpy()
    array([[0, 2, 1],
           [1, 2, 0],
           [2, 0, 1]], dtype=int32)
    >>> tf.argsort(D).numpy()
    array([[0, 2, 1],
           [1, 2, 0],
           [0, 2, 1]], dtype=int32)

    Notice that for tf.argsort on the third row, the first element is *not* 2.

    Args:
        D (tf.Tensor, np.ndarray): A matrix with distances (all entries should be nonnegative)

    Returns:
        tf.Tensor: a Tensor of the same shape as D, with indices that sort the rows
        (see tf.argsort).
    """
    return tf.argsort(D - 1.0 * tf.eye(tf.shape(D)[0], dtype=D.dtype))


@tf.function
def nearest_k(D, k) -> tuple[tf.Tensor, tf.Tensor]:
    return tf.math.top_k(-(D - tf.eye(tf.shape(D)[0], dtype=D.dtype)), k=k)


@tf.function
def psqdist(X) -> tf.Tensor:
    X = tf.cast(X, tf.float64)
    row_norms = tf.reshape(tf.reduce_sum(X**2, 1), [-1, 1])

    dists = row_norms - 2 * tf.matmul(X, X, transpose_b=True) + tf.reshape(row_norms, [1, -1])
    return tf.maximum(dists, 0.0)


@tf.function
def csqdist(X, Y) -> tf.Tensor:
    X = tf.cast(X, tf.float64)
    Y = tf.cast(Y, tf.float64)

    x_norms = tf.reshape(tf.reduce_sum(X**2, 1), [-1, 1])
    y_norms = tf.reshape(tf.reduce_sum(Y**2, 1), [1, -1])

    dists = x_norms - 2 * tf.matmul(X, Y, transpose_b=True) + y_norms
    return tf.maximum(dists, 0.0)


@tf.function
def flat_psqdist(X) -> tf.Tensor:
    D = psqdist(X)

    mask = tf.ones_like(D)
    mask = tf.linalg.band_part(mask, 0, -1) - tf.linalg.band_part(mask, 0, 0)
    mask = tf.cast(mask, bool)

    return tf.boolean_mask(D, mask)



from typing import Optional

import tensorflow as tf

from tensorflow_projection_qm.util import distance

from .metric import LocalizableMetric


@tf.function
def jaccard_impl(X, X_2d, k) -> tf.Tensor:
    k = tf.cast(k, tf.int32)
    D_high = distance.psqdist(X)
    D_low = distance.psqdist(X_2d)

    n = tf.shape(D_high)[0]
    k = tf.minimum(k, n - 1)

    knn_orig = distance.nearest_k(D_high, k=k + 1)[1][..., 1:]
    knn_proj = distance.nearest_k(D_low, k=k + 1)[1][..., 1:]

    intersection = tf.sets.intersection(knn_orig, knn_proj)
    union = tf.sets.union(knn_orig, knn_proj)

    n_intersection = tf.sparse.reduce_sum(tf.sparse.map_values(tf.ones_like, intersection), axis=-1)
    n_union = tf.sparse.reduce_sum(tf.sparse.map_values(tf.ones_like, union), axis=-1)

    return n_intersection / n_union
    # intersection_sizes = tf.sparse.reduce_count ?


def jaccard(X, X_2d, k):
    """Calculate the Jaccard quality metric for a projection

    Args:
        X (tf.Tensor, np.ndarray): the high-dimensional data matrix
        X_2d (tf.Tensor, np.ndarray): the projection data matrix
        k (tf.Tensor): a scalar tensor containing the number of neighbors to consider
        with_local (bool, optional): if true, returns both the scalar value of the metric as well
            as the value for each data point. Defaults to False.

    Returns:
        tf.Tensor: the scalar value of the Jaccard metric for the whole projection.
    """
    return tf.reduce_mean(jaccard_impl(X, X_2d, tf.constant(k)))


def jaccard_with_local(X, X_2d, k) -> tuple[tf.Tensor, tf.Tensor]:
    per_point = jaccard_impl(X, X_2d, tf.constant(k))
    return tf.reduce_mean(per_point), per_point


class Jaccard(LocalizableMetric):
    name = "jaccard"

    def __init__(self, k: Optional[int] = None) -> None:
        super().__init__()
        self.k = k

    @property
    def config(self):
        return {"k": self.k}

    def measure(self, X, X_2d):
        if self._with_local:
            return jaccard_with_local(X, X_2d, self.k)
        return jaccard(X, X_2d, self.k)

    def measure_from_dict(self, data_dict: dict[str, object]):
        return self.measure(data_dict["X"], data_dict["X_2d"])



from typing import Optional

import numpy as np
import tensorflow as tf

from tensorflow_projection_qm.metrics.metric import Metric


@tf.function
def distance_consistency_impl(X_2d, y, n_classes):
    y_sort_ixs = tf.argsort(y)

    X_2d = tf.gather(X_2d, y_sort_ixs)  # re-order, grouped per class
    y_sorted = tf.gather(y, y_sort_ixs)
    uniq, _, sizes = tf.unique_with_counts(y_sorted)
    per_class = tf.split(X_2d, sizes, num=n_classes)

    centroids = tf.stack(
        [tf.reduce_mean(single_class_data, axis=0) for single_class_data in per_class]
    )

    closest_centroid = tf.argmin(
        tf.linalg.norm(tf.expand_dims(X_2d, 1) - centroids, axis=-1), axis=1
    )
    closest_centroid = tf.gather(uniq, closest_centroid)

    return tf.reduce_mean(tf.cast(closest_centroid == y_sorted, tf.float64))


def distance_consistency(X_2d, y, n_classes):
    return distance_consistency_impl(X_2d, y, n_classes)


class DistanceConsistency(Metric):
    name = "distance_consistency"

    def __init__(self, n_classes: Optional[int] = None) -> None:
        super().__init__()
        self.n_classes = n_classes

    @property
    def config(self):
        return {"n_classes": self.n_classes}

    def measure(self, X_2d, y):
        return distance_consistency(X_2d, y, self.n_classes)

    def measure_from_dict(self, args: dict):
        return self.measure(args["X_2d"], args["y"])



from typing import Optional

import tensorflow as tf

from tensorflow_projection_qm.metrics.metric import LocalizableMetric
from tensorflow_projection_qm.util import distance


@tf.function
def false_neighbors_impl(X, X_2d, k):
    k = tf.cast(k, tf.int32)

    D_high = distance.psqdist(X)
    D_low = distance.psqdist(X_2d)

    _, knn_orig = distance.nearest_k(D_high, k)
    _, knn_proj = distance.nearest_k(D_low, k)

    false_neighbors = tf.sparse.reduce_sum(
        tf.sparse.map_values(tf.ones_like, tf.sets.difference(knn_proj, knn_orig)), axis=-1
    )
    false_neighbors /= k

    return false_neighbors


@tf.function
def missing_neighbors_impl(X, X_2d, k):
    k = tf.cast(k, tf.int32)

    D_high = distance.psqdist(X)
    D_low = distance.psqdist(X_2d)

    _, knn_orig = distance.nearest_k(D_high, k)
    _, knn_proj = distance.nearest_k(D_low, k)

    missing_neighbors = tf.sparse.reduce_sum(
        tf.sparse.map_values(tf.ones_like, tf.sets.difference(knn_orig, knn_proj)), axis=-1
    )
    missing_neighbors /= k

    return missing_neighbors


@tf.function
def true_neighbors_impl(X, X_2d, k):
    k = tf.cast(k, tf.int32)

    D_high = distance.psqdist(X)
    D_low = distance.psqdist(X_2d)

    _, knn_orig = distance.nearest_k(D_high, k)
    _, knn_proj = distance.nearest_k(D_low, k)

    true_neighbors = tf.sparse.reduce_sum(
        tf.sparse.map_values(tf.ones_like, tf.sets.intersection(knn_orig, knn_proj)), axis=-1
    )
    true_neighbors /= k
    return true_neighbors


def false_neighbors(X, X_2d, k):
    return tf.reduce_mean(false_neighbors_impl(X, X_2d, tf.constant(k)))


def false_neighbors_with_local(X, X_2d, k):
    per_point = false_neighbors_impl(X, X_2d, tf.constant(k))
    return tf.reduce_mean(per_point), per_point


def missing_neighbors(X, X_2d, k):
    return tf.reduce_mean(missing_neighbors_impl(X, X_2d, tf.constant(k)))


def missing_neighbors_with_local(X, X_2d, k):
    per_point = missing_neighbors_impl(X, X_2d, tf.constant(k))
    return tf.reduce_mean(per_point), per_point


def true_neighbors(X, X_2d, k):
    return tf.reduce_mean(true_neighbors_impl(X, X_2d, tf.constant(k)))


def true_neighbors_with_local(X, X_2d, k):
    per_point = true_neighbors_impl(X, X_2d, tf.constant(k))
    return tf.reduce_mean(per_point), per_point


class FalseNeighbors(LocalizableMetric):
    name = "false_neighbors"

    def __init__(self, k: Optional[int] = None) -> None:
        super().__init__()
        self.k = k

    @property
    def config(self):
        return {"k": self.k}

    def measure(self, X, X_2d):
        if self._with_local:
            return false_neighbors_with_local(X, X_2d, self.k)
        return false_neighbors(X, X_2d, self.k)

    def measure_from_dict(self, args: dict):
        return self.measure(args["X"], args["X_2d"])


class MissingNeighbors(LocalizableMetric):
    name = "missing_neighbors"

    def __init__(self, k: Optional[int] = None) -> None:
        super().__init__()
        self.k = k

    @property
    def config(self):
        return {"k": self.k}

    def measure(self, X, X_2d):
        if self._with_local:
            return missing_neighbors_with_local(X, X_2d, self.k)
        return missing_neighbors(X, X_2d, self.k)

    def measure_from_dict(self, args: dict):
        return self.measure(args["X"], args["X_2d"])


class TrueNeighbors(LocalizableMetric):
    name = "true_neighbors"

    def __init__(self, k: Optional[int] = None) -> None:
        super().__init__()
        self.k = k

    @property
    def config(self):
        return {"k": self.k}

    def measure(self, X, X_2d):
        if self._with_local:
            return true_neighbors_with_local(X, X_2d, self.k)
        return true_neighbors(X, X_2d, self.k)

    def measure_from_dict(self, args: dict):
        return self.measure(args["X"], args["X_2d"])



from typing import Optional

import tensorflow as tf

from tensorflow_projection_qm.metrics.metric import LocalizableMetric
from tensorflow_projection_qm.util import distance


@tf.function
def continuity_impl(X, X_2d, k) -> tf.Tensor:
    k = tf.cast(k, tf.int32)
    D_high = distance.psqdist(X)
    D_low = distance.psqdist(X_2d)

    n = tf.shape(D_high)[0]
    k = tf.minimum(k, n - 1)

    if 2 * k < n:
        norm_factor = k * (2 * n - 3 * k - 1) / 2
    else:
        norm_factor = (n - k) * (n - k - 1) / 2

    knn_orig = distance.nearest_k(D_high, k=k + 1)[1][:, 1:]
    nn_proj = distance.sort_distances(D_low)
    ixs_proj = tf.argsort(nn_proj)

    knn_proj = nn_proj[:, 1 : k + 1]

    V_i = tf.sparse.to_dense(tf.sets.difference(knn_orig, knn_proj), default_value=-1)
    pre_cont = tf.where(
        V_i >= 0, tf.gather(ixs_proj, tf.where(V_i >= 0, V_i, 0), batch_dims=-1) - k, 0
    )
    cont = tf.reduce_sum(pre_cont, -1)
    cont_t = tf.cast(cont, tf.float64)
    k = tf.cast(k, tf.float64)
    n = tf.cast(n, tf.float64)

    return tf.squeeze(1 - tf.math.divide_no_nan(cont_t, norm_factor))


@tf.function
def class_aware_continuity_impl(X, X_2d, y, k, n_classes):
    k = tf.cast(k, tf.int32)
    D_high = distance.psqdist(X)
    D_low = distance.psqdist(X_2d)

    n = tf.shape(D_high)[0]
    k = tf.minimum(k, n - 1)
    if 2 * k < n:
        norm_factor = k * (2 * n - 3 * k - 1) / 2
    else:
        norm_factor = (n - k) * (n - k - 1) / 2

    nn_proj = distance.sort_distances(D_low)
    ixs_proj = tf.argsort(nn_proj)

    knn_orig = distance.nearest_k(D_high, k=k + 1)[1][:, 1:]
    knn_proj = nn_proj[:, 1 : k + 1]

    missing = tf.sparse.to_dense(tf.sets.difference(knn_orig, knn_proj), default_value=-1)
    classes = tf.where(missing >= 0, tf.gather(y, tf.where(missing >= 0, missing, 0)), -1)

    V_i = tf.where(classes == y[:, None], missing, -1)
    pre_cont = tf.where(
        V_i >= 0, tf.gather(ixs_proj, tf.where(V_i >= 0, V_i, 0), batch_dims=-1) - k, 0
    )
    cont = tf.reduce_sum(pre_cont, -1)
    cont_t = tf.cast(cont, tf.float64)
    k = tf.cast(k, tf.float64)
    n = tf.cast(n, tf.float64)

    return tf.squeeze(1 - tf.math.multiply_no_nan(1 / norm_factor, cont_t))


def continuity(X, X_2d, k: int) -> tf.Tensor:
    return tf.reduce_mean(continuity_impl(X, X_2d, tf.constant(k)))


def continuity_with_local(X, X_2d, k: int) -> tuple[tf.Tensor, tf.Tensor]:
    per_point = continuity_impl(X, X_2d, tf.constant(k))
    return tf.reduce_mean(per_point), per_point


def class_aware_continuity(X, X_2d, y, k, n_classes):
    return tf.reduce_mean(
        class_aware_continuity_impl(X, X_2d, y, tf.constant(k), tf.constant(n_classes))
    )


def class_aware_continuity_with_local(X, X_2d, y, k, n_classes):
    per_point = class_aware_continuity_impl(X, X_2d, y, tf.constant(k), tf.constant(n_classes))
    return tf.reduce_mean(per_point), per_point


class Continuity(LocalizableMetric):
    name = "continuity"

    def __init__(self, k: Optional[int] = None) -> None:
        super().__init__()
        self.k = k

    @property
    def config(self):
        return {"k": self.k}

    def measure(self, X, X_2d):
        if self._with_local:
            return continuity_with_local(X, X_2d, self.k)
        return continuity(X, X_2d, self.k)

    def measure_from_dict(self, args: dict):
        return self.measure(args["X"], args["X_2d"])


class ClassAwareContinuity(LocalizableMetric):
    name = "class_aware_continuity"

    def __init__(self, k: Optional[int] = None, n_classes: Optional[int] = None) -> None:
        super().__init__()
        self.k = k
        self.n_classes = n_classes

    @property
    def config(self):
        return {"k": self.k, "n_classes": self.n_classes}

    def measure(self, X, X_2d, y):
        if self._with_local:
            return class_aware_continuity_with_local(X, X_2d, y, self.k, self.n_classes)
        return class_aware_continuity(X, X_2d, y, self.k, self.n_classes)

    def measure_from_dict(self, args: dict):
        return self.measure(args["X"], args["X_2d"], args["y"])



from abc import ABC, abstractmethod
from typing import Iterable, TypeVar

import tensorflow as tf

# typing.Self only available in Python>=3.11.
TLocalizableMetric = TypeVar("TLocalizableMetric", bound="LocalizableMetric")
TMetric = TypeVar("TMetric", bound="Metric")


class Metric(ABC):
    name: str

    def __init__(self) -> None:
        super().__init__()

    @property
    @abstractmethod
    def config(self) -> dict: ...

    @abstractmethod
    def measure(self, *args, **kwargs): ...

    @abstractmethod
    def measure_from_dict(self, args: dict): ...

    def set_if_missing(self: TMetric, params) -> TMetric:
        new = type(self)(**self.config)
        for k in new.config:
            if k in params and getattr(self, k) is None:
                setattr(new, k, params[k])
        return new


class LocalizableMetric(Metric):
    def __init__(self) -> None:
        super().__init__()
        self._with_local = False

    def with_local(self: TLocalizableMetric) -> TLocalizableMetric:
        new = type(self)(**self.config)
        new._with_local = True
        return new


class MetricSet:
    def __init__(self, metrics: Iterable[Metric], defaults: dict = {}) -> None:
        self.metrics = list(metrics)
        self.defaults: dict = {} | defaults

    def set_default(self, **kwargs):
        self.defaults |= kwargs

    @tf.function
    def _measure(self, X, X_2d, y=None):
        return {
            m.name: m.set_if_missing(self.defaults).measure_from_dict(
                {"X": X, "X_2d": X_2d, "y": y}
            )
            for m in self.metrics
        }

    def measure_from_dict(self, data_dict: dict):
        # We split the dict and call the underlying measure_from_dict
        # so that we can annotate _measure with @tf.function without
        # causing frequent retracing.
        return self._measure(data_dict["X"], data_dict["X_2d"], data_dict.get("y"))



from . import (
    avg_local_error,
    continuity,
    distance_consistency,
    jaccard,
    mean_rel_rank_error,
    metric,
    neighborhood_hit,
    neighbors,
    pearson_correlation,
    shepard_goodness,
    stress,
    trustworthiness,
)

_ALL_LOCALIZABLE_METRICS: tuple[metric.LocalizableMetric, ...] = (
    avg_local_error.AverageLocalError(),
    continuity.ClassAwareContinuity(),
    continuity.Continuity(),
    jaccard.Jaccard(),
    mean_rel_rank_error.MRREData(),
    mean_rel_rank_error.MRREProj(),
    neighborhood_hit.NeighborhoodHit(),
    neighbors.FalseNeighbors(),
    neighbors.MissingNeighbors(),
    neighbors.TrueNeighbors(),
    trustworthiness.ClassAwareTrustworthiness(),
    trustworthiness.Trustworthiness(),
)
_ALL_METRICS: tuple[metric.Metric, ...] = _ALL_LOCALIZABLE_METRICS + (
    distance_consistency.DistanceConsistency(),
    pearson_correlation.PearsonCorrelation(),
    shepard_goodness.ShepardGoodness(),
    stress.NormalizedStress(),
    stress.ScaleNormalizedStress(),
)
_ALL_METRICS_METRICSET = metric.MetricSet(list(_ALL_METRICS))


def get_all_metrics_runner(defaults: dict) -> metric.MetricSet:
    ms = metric.MetricSet(list(_ALL_METRICS))
    ms.set_default(**defaults)

    return ms


def run_all_metrics(X, X_2d, y, k, n_classes=None, *, as_numpy=False):
    _ALL_METRICS_METRICSET.set_default(k=k, n_classes=n_classes)
    measures = _ALL_METRICS_METRICSET.measure_from_dict({"X": X, "X_2d": X_2d, "y": y})

    if as_numpy:
        measures = {k: v.numpy() for k, v in measures.items()}
    return measures
